"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.StreamBuffer = exports.ChunkBuffer = void 0;
class ChunkBuffer {
    constructor(name, bytes, sbuf, buffer = Buffer.alloc(bytes), ofs = 0) {
        this.name = name;
        this.bytes = bytes;
        this.sbuf = sbuf;
        this.buffer = buffer;
        this.ofs = ofs;
    }
    writeBoolean(val) {
        // console.log('writeBoolean:', val);
        this.buffer.writeUInt8(~~val, this.ofs);
        this.ofs += 1;
    }
    writeUint8(val) {
        this.buffer.writeUInt8(~~val, this.ofs);
        this.ofs += 1;
    }
    writeChar(val) {
        this.writeUint8(val);
    }
    writeUint16(val) {
        this.buffer.writeUInt16LE(val, this.ofs);
        this.ofs += 2;
    }
    writeShort(val) {
        this.buffer.writeInt16LE(val, this.ofs);
        this.ofs += 2;
    }
    writeUint32(val) {
        this.buffer.writeUInt32LE(val, this.ofs);
        this.ofs += 4;
    }
    writeInt(val) {
        // console.log('writeInt', val, ~~val, this.ofs);
        this.buffer.writeInt32LE(~~val, this.ofs);
        this.ofs += 4;
    }
    writeFloat(val) {
        this.buffer.writeFloatLE(val, this.ofs);
        this.ofs += 4;
    }
    // public writeUint64(val: HighLow) {
    //   HighLow(val, this.sbuf);
    // }
    // public writeLong(val: HighLow) {
    //   HighLowToStream(val, this.sbuf);
    // }
    writeDouble(val) {
        this.buffer.writeDoubleLE(val, this.ofs);
        this.ofs += 8;
    }
    readBoolean() {
        const ret = this.buffer.readUInt8(this.ofs);
        this.ofs += 1;
        return !!ret;
    }
    readUint8() {
        const ret = this.buffer.readUInt8(this.ofs);
        this.ofs += 1;
        return ret;
    }
    readChar() {
        return this.readUint8();
    }
    readUint16() {
        const ret = this.buffer.readUInt16LE(this.ofs);
        this.ofs += 2;
        return ret;
    }
    readShort() {
        const ret = this.buffer.readInt16LE(this.ofs);
        this.ofs += 2;
        return ret;
    }
    readUint32() {
        const ret = this.buffer.readUInt32LE(this.ofs);
        this.ofs += 4;
        return ret;
    }
    readInt() {
        const ret = this.buffer.readUInt32LE(this.ofs);
        // console.log('readInt', ret, this.ofs);
        this.ofs += 4;
        return ret;
    }
    readFloat() {
        const ret = this.buffer.readFloatLE(this.ofs);
        this.ofs += 4;
        return ret;
    }
    readDouble() {
        const ret = this.buffer.readDoubleLE(this.ofs);
        this.ofs += 8;
        return ret;
    }
}
exports.ChunkBuffer = ChunkBuffer;
class StreamBuffer {
    constructor(u8s = []) {
        this.buffers = [];
        const totalLen = u8s.reduce((r, u8) => r + u8.length, 0);
        const buf = u8s.reduce((b, u8) => {
            // console.log(b.ofs, u8);
            b.buf.set(u8, b.ofs);
            b.ofs += u8.length;
            return b;
        }, { buf: Buffer.alloc(totalLen), ofs: 0 });
        if (totalLen > 0) {
            // console.log('Read:', totalLen, buf.buf);
            this.buffers.push(new ChunkBuffer('reader', totalLen, this, buf.buf, 0));
        }
    }
    currentWriteChunk(name, bytes) {
        let lastPos = this.buffers.length - 1;
        let last;
        if (lastPos >= 0) {
            last = this.buffers[lastPos];
            if (last.ofs + bytes > last.bytes) {
                // not nice but fast
                lastPos = -1;
            }
        }
        if (lastPos < 0) {
            last = new ChunkBuffer(name, bytes, this);
            this.buffers.push(last);
        }
        return last;
    }
    currentReadChunk(name, bytes) {
        const c = this.buffers[this.buffers.length - 1];
        if (c.ofs + bytes > c.bytes) {
            throw Error(`currentReadChunk: ${name}-${bytes}`);
        }
        return c;
    }
    prepareRead(name, bytes, cb) {
        return cb(this.currentReadChunk(name, bytes));
    }
    prepareWrite(name, bytes, cb) {
        const buffer = this.currentWriteChunk(name, bytes);
        cb(buffer);
        return this;
    }
    asUint8Array() {
        const totalLen = this.buffers.reduce((r, i) => r + i.buffer.length, 0);
        const out = new Uint8Array(totalLen);
        this.buffers.reduce((r, b) => {
            out.set(b.buffer, r);
            return r + b.buffer.length;
        }, 0);
        return out;
    }
}
exports.StreamBuffer = StreamBuffer;
//# sourceMappingURL=stream-buffer.js.map