import { Funcs } from '../align';
import { Option } from '../optional';
import { Definition as Uint32 } from './uint32';
import { Definition as Base, TypeName, NamedType } from './base';
import { ChunkBuffer } from '../stream-buffer';
export interface BitItem {
    readonly name: string;
    readonly start: number;
    readonly length?: number;
    readonly initial?: number | boolean;
}
export interface BooleanBitItemWithLength extends BitItem {
    readonly length: 1;
    readonly type: Boolean;
}
export interface UInt32BitItemWithLength extends BitItem {
    readonly length: number;
    readonly type: Uint32;
}
export interface BitItemWithLength extends BitItem {
    readonly length: number;
    readonly type: Base<boolean | number>;
}
export declare type BitStructInitial = Record<string, boolean | number>;
export interface BitStructArg {
    readonly name: string;
    readonly length: number;
    readonly bits: BitItem[];
    readonly alignFuncs: Funcs<string>;
    readonly initial: BitStructInitial;
}
export declare type BitsByName = Record<string, BitItemWithLength>;
export declare function BitStructValue(val: number | boolean, bits: number): number | boolean;
export declare const typeName: TypeName;
export declare abstract class AbstractDefinition extends NamedType<BitStructInitial> {
    readonly type: TypeName;
    abstract readonly bytes: number;
    abstract readonly length: number;
    abstract readonly alignFuncs: Funcs<string>;
    abstract readonly bits: BitItemWithLength[];
    abstract readonly bitsByName: BitsByName;
    abstract readonly givenInitial: Option<Partial<BitStructInitial>>;
}
export declare class Definition extends AbstractDefinition {
    static readonly type: TypeName;
    readonly name: string;
    readonly bytes: number;
    readonly length: number;
    readonly alignFuncs: Funcs<string>;
    readonly bits: BitItemWithLength[];
    readonly bitsByName: BitsByName;
    readonly givenInitial: Option<Partial<BitStructInitial>>;
    constructor(arg: Partial<BitStructArg>);
    fromStreamChunk(chunk: ChunkBuffer, name?: TypeName): Record<string, number | boolean>;
    toStreamChunk(val: Record<string, number | boolean>, chunk: ChunkBuffer, name?: TypeName): void;
    create(...rargs: Partial<BitStructInitial>[]): BitStructInitial;
    coerce(val?: BitStructInitial): Option<BitStructInitial>;
}
export declare type BitStruct = Definition;
