"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Definition = exports.AbstractDefinition = exports.typeName = exports.BitStructValue = void 0;
const align_1 = require("../align");
const optional_1 = require("../optional");
const boolean_1 = require("./boolean");
const uint32_1 = require("./uint32");
const base_1 = require("./base");
function BitStructValue(val, bits) {
    if (bits === 1) {
        return !!val;
    }
    return ~~val & (2 ** bits - 1);
}
exports.BitStructValue = BitStructValue;
exports.typeName = 'BitStruct';
class AbstractDefinition extends base_1.NamedType {
    constructor() {
        super(...arguments);
        this.type = exports.typeName;
    }
}
exports.AbstractDefinition = AbstractDefinition;
let Definition = /** @class */ (() => {
    class Definition extends AbstractDefinition {
        constructor(arg) {
            super();
            const al = align_1.funcsMapper({ ...arg.alignFuncs, element: 'byte' });
            this.alignFuncs = al.names;
            this.length = typeof arg.length === 'number' ? arg.length : 1;
            this.bytes = al.funcs.overall(this.length);
            const tmpLen = new Set((arg.bits || []).map((i) => i.name));
            if (tmpLen.size != (arg.bits || []).length) {
                throw Error('The bits are double definied');
            }
            this.bits = (arg.bits || []).map((bit) => {
                const length = typeof bit.length === 'number' ? bit.length : 1;
                if (bit.start + length > this.bytes * 8) {
                    throw RangeError(`BitStruct:${bit.name} requesting more bits than length:${this.bytes}:${bit.start}:${length}`);
                }
                const my = { ...bit };
                delete my.initial;
                if (length === 1) {
                    return {
                        ...my,
                        length: 1,
                        type: new boolean_1.Definition(!!bit.initial ? { initial: !!bit.initial } : undefined),
                    };
                }
                else {
                    return {
                        ...my,
                        length,
                        type: new uint32_1.Definition(~~bit.initial ? { initial: ~~bit.initial } : undefined),
                    };
                }
            });
            this.bitsByName = this.bits.reduce((r, b) => {
                r[b.name] = b;
                return r;
            }, {});
            this.name =
                typeof arg.name === 'string'
                    ? arg.name
                    : (this.bits || [])
                        .map((i) => `${i.name}S${i.start}L${i.length}` +
                        `${['boolean', 'number'].includes(typeof i.initial) ? `I${~~i.initial}` : ''}`)
                        .sort((a, b) => a.localeCompare(b))
                        .join('_');
            this.givenInitial = this.coerce(arg.initial);
        }
        fromStreamChunk(chunk, name = this.type) {
            throw new Error('Method not implemented.');
        }
        toStreamChunk(val, chunk, name = this.type) {
            throw new Error('Method not implemented.');
        }
        create(...rargs) {
            const data = rargs
                .concat(optional_1.OrUndefined(this.givenInitial))
                .filter((i) => typeof i === 'object')
                .reduce((r, val) => {
                this.bits.forEach((bit) => {
                    if (['boolean', 'number'].includes(typeof val[bit.name])) {
                        if (bit.length === 1) {
                            r[bit.name].push(!!val[bit.name]);
                        }
                        else {
                            r[bit.name].push(~~val[bit.name]);
                        }
                    }
                });
                return r;
            }, this.bits.reduce((r, bit) => {
                r[bit.name] = [];
                return r;
            }, {}));
            return this.bits.reduce((r, bit) => {
                r[bit.name] = bit.type.create(...data[bit.name]);
                return r;
            }, {});
        }
        coerce(val) {
            let ret = optional_1.NoneOption;
            if (typeof val === 'object') {
                this.bits.forEach((bit) => {
                    if (['boolean', 'number'].includes(typeof val[bit.name])) {
                        if (optional_1.isNone(ret)) {
                            ret = optional_1.SomeOption({});
                        }
                        ret.some[bit.name] = bit.length === 1 ? !!val[bit.name] : val[bit.name];
                    }
                });
            }
            return ret;
        }
    }
    Definition.type = exports.typeName;
    return Definition;
})();
exports.Definition = Definition;
//# sourceMappingURL=bit-struct.js.map