"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Definition = exports.create = exports.ArrayTypeAttribute = void 0;
const optional_1 = require("../optional");
const base_1 = require("./base");
const align_1 = require("../align");
class ArrayTypeAttribute extends base_1.Definition {
}
exports.ArrayTypeAttribute = ArrayTypeAttribute;
function create(len, cb) {
    const my = Array(len);
    for (let i = 0; i < len; ++i) {
        my[i] = cb(i);
    }
    return my;
}
exports.create = create;
// type fromStreamFN<B> = (rb: StreamBuffer, name: string) => B[];
// type toStreamFN<B> = (_val: B[], wb: StreamBuffer, name: string) => void;
let Definition = /** @class */ (() => {
    class Definition extends ArrayTypeAttribute {
        constructor(el) {
            super();
            this.type = Definition.type;
            this.length = el.length;
            const al = align_1.funcsMapper(el.alignFuncs);
            this.alignFuncs = al.names;
            this.bytes = al.funcs.overall(el.length * al.funcs.element(el.element.bytes));
            this.element = el.element;
            if (el.initial) {
                this.givenInitial = optional_1.SomeOption(el.initial);
            }
            else {
                this.givenInitial = optional_1.NoneOption;
            }
        }
        create(...initials) {
            const datas = initials
                .concat([
                optional_1.OrUndefined(this.givenInitial),
                new Array(this.length).fill(this.element.create()),
            ])
                .filter((i) => Array.isArray(i));
            const items = datas.reduce((r, bArray) => {
                bArray.slice(0, this.length).forEach((item, idx) => {
                    const v = this.element.coerce(item);
                    if (optional_1.isSome(v)) {
                        r[idx].push(v.some);
                    }
                });
                return r;
            }, new Array(this.length).fill(undefined).map((i) => []));
            return items.reduce((r, item, idx) => {
                r[idx] = this.element.create(...item);
                return r;
            }, new Array(this.length));
        }
        coerce(vals) {
            if (!Array.isArray(vals)) {
                return optional_1.NoneOption;
            }
            let found = false;
            const ret = vals.map((i) => {
                const e = this.element.coerce(i);
                if (optional_1.isNone(e)) {
                    return undefined;
                }
                found = true;
                return e.some;
            });
            if (!found) {
                return optional_1.NoneOption;
            }
            return optional_1.SomeOption(ret);
        }
        fromStreamChunk(nrb, name = this.type) {
            const ret = Array(this.length);
            for (let i = 0; i < this.length; ++i) {
                ret[i] = this.element.fromStreamChunk(nrb, `${name}.${i}`);
            }
            return ret;
        }
        toStreamChunk(val, nwb, name = this.type) {
            for (let i = 0; i < this.length; ++i) {
                this.element.toStreamChunk(val[i], nwb, `${name}.${i}`);
            }
        }
    }
    Definition.type = 'FixedArray';
    return Definition;
})();
exports.Definition = Definition;
//# sourceMappingURL=fixed-array.js.map