"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Definition = void 0;
const optional_1 = require("../optional");
const base_1 = require("./base");
const char_1 = require("./char");
const align_1 = require("../align");
let Definition = /** @class */ (() => {
    class Definition extends base_1.Definition {
        constructor(iel) {
            super();
            this.type = Definition.type;
            this.element = Definition.element;
            const el = iel; // artefact
            const al = align_1.funcsMapper({ ...el.alignFuncs, element: 'byte' });
            this.alignFuncs = al.names;
            this.length = el.length;
            this.bytes = al.funcs.overall(el.length);
            this.givenInitial = this.coerce(el.initial);
        }
        create(...initials) {
            const gi = optional_1.OrUndefined(this.givenInitial);
            const datas = initials
                .concat(gi ? [gi] : undefined)
                .map((i) => this.coerce(i))
                .filter((i) => optional_1.isSome(i))
                .map((i) => optional_1.isSome(i) && i.some)
                .concat([new Array(this.length).fill(Definition.element.create())]);
            const items = datas.reduce((r, bArray) => {
                bArray.forEach((item, idx) => {
                    r[idx].push(item);
                });
                return r;
            }, new Array(this.length).fill(undefined).map((i) => []));
            return items.reduce((r, item, idx) => {
                r[idx] = Definition.element.create(...item);
                return r;
            }, new Array(this.length).fill(Definition.element.create()));
        }
        coerce(m) {
            if (typeof m === 'string' || Array.isArray(m)) {
                const r = Array.from(m.slice(0, this.length - 1)).map((item) => {
                    const x = Definition.element.coerce(item);
                    return optional_1.isSome(x) ? x.some : 0;
                });
                if (r.length === this.length - 1) {
                    r.push(0); // trailing 0
                }
                return optional_1.SomeOption(r);
            }
            return optional_1.NoneOption;
        }
        fromStreamChunk(chunk, name = this.type) {
            const ret = Array(this.length);
            for (let i = 0; i < this.length; ++i) {
                const val = chunk.readUint8();
                ret[i] = val;
            }
            return ret;
        }
        toStreamChunk(val, chunk, name = this.type) {
            for (let i = 0; i < this.length; ++i) {
                chunk.writeUint8(val[i] || 0);
            }
        }
    }
    Definition.type = 'FixedCString';
    Definition.element = new char_1.Definition();
    return Definition;
})();
exports.Definition = Definition;
//# sourceMappingURL=fixed-cstring.js.map