"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HighLowType = void 0;
const optional_1 = require("../optional");
const base_1 = require("./base");
const uint32_1 = require("./uint32");
let HighLowType = /** @class */ (() => {
    class HighLowType extends base_1.Definition {
        constructor(ival) {
            super();
            this.bytes = HighLowType.bytes;
            const val = (ival || {}).initial;
            this.givenInitial = this.coerce(val);
        }
        coerce(hl) {
            if (typeof hl === 'object') {
                const high = HighLowType.uint32.coerce(hl.high);
                const low = HighLowType.uint32.coerce(hl.low);
                if (optional_1.isSome(high) && optional_1.isSome(low)) {
                    return optional_1.SomeOption({ low: low.some, high: high.some });
                }
                if (optional_1.isSome(high)) {
                    return optional_1.SomeOption({ high: high.some });
                }
                if (optional_1.isSome(low)) {
                    return optional_1.SomeOption({ low: low.some });
                }
            }
            return optional_1.NoneOption;
        }
        create(...args) {
            const data = args
                .concat(optional_1.OrUndefined(this.givenInitial))
                .concat({ high: 0, low: 0 })
                .filter((i) => typeof i === 'object')
                .reduce((r, i) => {
                const v = this.coerce(i);
                if (optional_1.isSome(v)) {
                    r.push(v.some);
                }
                return r;
            }, []);
            return Object.assign({}, ...data.reverse());
        }
        fromStreamChunk(chunk, name = 'HighLow') {
            return {
                low: chunk.readUint32(),
                high: chunk.readUint32(),
            };
        }
        toStreamChunk(val, chunk, name = 'HighLow') {
            chunk.writeUint32(val.low);
            chunk.writeUint32(val.high);
        }
    }
    HighLowType.uint32 = new uint32_1.Definition();
    HighLowType.bytes = 8;
    return HighLowType;
})();
exports.HighLowType = HighLowType;
//# sourceMappingURL=high-low.js.map