"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.toAttributeType = exports.AttributeType = exports.isStruct = exports.isFixedArray = exports.isHighLow = exports.isFixedCString = exports.isBitStruct = exports.isScalar = exports.ScalarTypesList = exports.SimpleScalarTypesList = void 0;
exports.Base = require("./base");
exports.Boolean = require("./boolean");
exports.Char = require("./char");
exports.Uint8 = require("./uint8");
exports.Uint16 = require("./uint16");
exports.Short = require("./short");
exports.Uint32 = require("./uint32");
exports.Int = require("./int");
exports.Float = require("./float");
exports.Double = require("./double");
exports.Long = require("./long");
exports.Uint64 = require("./uint64");
exports.FixedCString = require("./fixed-cstring");
exports.FixedArray = require("./fixed-array");
exports.BitStruct = require("./bit-struct");
exports.Struct = require("./struct");
const boolean_1 = require("./boolean");
const uint8_1 = require("./uint8");
const uint16_1 = require("./uint16");
const uint32_1 = require("./uint32");
const char_1 = require("./char");
const short_1 = require("./short");
const int_1 = require("./int");
const double_1 = require("./double");
const float_1 = require("./float");
const uint64_1 = require("./uint64");
const long_1 = require("./long");
const fixed_array_1 = require("./fixed-array");
const fixed_cstring_1 = require("./fixed-cstring");
const bit_struct_1 = require("./bit-struct");
const struct_1 = require("./struct");
exports.SimpleScalarTypesList = [
    boolean_1.Definition,
    uint8_1.Definition,
    char_1.Definition,
    uint16_1.Definition,
    short_1.Definition,
    uint32_1.Definition,
    int_1.Definition,
    float_1.Definition,
    uint64_1.Definition,
    long_1.Definition,
    double_1.Definition,
];
exports.ScalarTypesList = [...exports.SimpleScalarTypesList, fixed_cstring_1.Definition, bit_struct_1.Definition];
function isScalar(def) {
    return !!exports.ScalarTypesList.find((i) => def.type === i.type);
}
exports.isScalar = isScalar;
function isBitStruct(def) {
    return def.type === bit_struct_1.Definition.type;
}
exports.isBitStruct = isBitStruct;
function isFixedCString(def) {
    return def.type === fixed_cstring_1.Definition.type;
}
exports.isFixedCString = isFixedCString;
function isHighLow(def) {
    return def.type === long_1.Definition.type || def.type === uint64_1.Definition.type;
}
exports.isHighLow = isHighLow;
function isFixedArray(def) {
    return def.type === fixed_array_1.Definition.type;
}
exports.isFixedArray = isFixedArray;
function isStruct(def) {
    return def.type === struct_1.Definition.type;
}
exports.isStruct = isStruct;
var AttributeType;
(function (AttributeType) {
    AttributeType[AttributeType["Scalar"] = 0] = "Scalar";
    AttributeType[AttributeType["FixedArray"] = 1] = "FixedArray";
    AttributeType[AttributeType["Struct"] = 2] = "Struct";
})(AttributeType = exports.AttributeType || (exports.AttributeType = {}));
function toAttributeType(def) {
    if (isScalar(def)) {
        return AttributeType.Scalar;
    }
    if (isFixedArray(def)) {
        return AttributeType.FixedArray;
    }
    if (isStruct(def)) {
        return AttributeType.Struct;
    }
    throw Error(`Unknown type: ${def.type}`);
}
exports.toAttributeType = toAttributeType;
//# sourceMappingURL=index.js.map