import { Option } from '../optional';
import { Definition as Base, TypeName, NamedType } from './base';
import { Funcs } from '../align';
import { ChunkBuffer } from '../stream-buffer';
export interface BaseAttribute {
    readonly name: string;
}
export interface Attribute<T> extends BaseAttribute {
    readonly type: Base<T>;
}
export interface AttributeOfs<T> extends Attribute<T> {
    readonly ofs: number;
}
export declare type StructInitial = Record<string, any>;
export declare type StructByName = Record<string, AttributeOfs<any>>;
export declare type StructMerge = Record<string, unknown[]>;
export interface StructArg {
    readonly name: string;
    readonly alignFuncs?: Partial<Funcs<string>>;
    readonly attributes: Attribute<unknown>[];
    readonly initial?: StructInitial;
}
export declare const typeName: TypeName;
export declare abstract class AbstractDefinition extends NamedType<StructInitial> {
    readonly type: TypeName;
    abstract readonly bytes: number;
    abstract readonly name: string;
    abstract readonly alignFuncs: Funcs<string>;
    abstract readonly attributes: AttributeOfs<any>[];
    abstract readonly attributeByName: StructByName;
    abstract readonly givenInitial: Option<Partial<StructInitial>>;
}
export declare class Definition extends AbstractDefinition {
    static readonly type: TypeName;
    readonly bytes: number;
    readonly name: string;
    readonly alignFuncs: Funcs<string>;
    readonly attributes: AttributeOfs<any>[];
    readonly attributeByName: StructByName;
    readonly givenInitial: Option<Partial<StructInitial>>;
    constructor(st: StructArg);
    coerce(vals: Record<string, any>): Option<Record<string, any>>;
    create(...rargs: StructInitial[]): StructInitial;
    fromStreamChunk: (chunk: ChunkBuffer, name?: string) => Record<string, any>;
    toStreamChunk(val: Record<string, any>, chunk: ChunkBuffer): void;
}
export declare type Struct = Definition;
