"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Definition = exports.AbstractDefinition = exports.typeName = void 0;
const optional_1 = require("../optional");
const base_1 = require("./base");
const align_1 = require("../align");
exports.typeName = 'Struct';
class AbstractDefinition extends base_1.NamedType {
    constructor() {
        super(...arguments);
        this.type = exports.typeName;
    }
}
exports.AbstractDefinition = AbstractDefinition;
let Definition = /** @class */ (() => {
    class Definition extends AbstractDefinition {
        // Type(typelevel)
        constructor(st) {
            super();
            // we need this defined in the class not in the prototype
            // tslint:disable-next-line: typedef
            this.fromStreamChunk = function (chunk, name = this.type) {
                throw new Error('Method not implemented.');
            };
            this.name = st.name;
            const al = align_1.funcsMapper(st.alignFuncs);
            this.alignFuncs = al.names;
            const tmp = st.attributes.reduce((res, attr) => {
                // Hack
                res.attributesInclOfs.push({
                    ...attr,
                    // notRequired: !!attr.notRequired,
                    ofs: res.bytes,
                });
                res.bytes = res.bytes + al.funcs.element(attr.type.bytes);
                return res;
            }, {
                bytes: 0,
                attributesInclOfs: [],
            });
            this.bytes = al.funcs.overall(tmp.bytes);
            this.attributes = tmp.attributesInclOfs;
            this.attributeByName = this.attributes.reduce((r, attr) => {
                r[attr.name] = attr;
                return r;
            }, {});
            this.givenInitial = this.coerce(st.initial);
        }
        coerce(vals) {
            let ret = optional_1.NoneOption;
            if (typeof vals === 'object') {
                this.attributes.forEach((attr) => {
                    const val = attr.type.coerce(vals[attr.name]);
                    if (optional_1.isSome(val)) {
                        if (optional_1.isNone(ret)) {
                            ret = optional_1.SomeOption({});
                        }
                        ret.some[attr.name] = val.some;
                    }
                });
            }
            return ret;
        }
        create(...rargs) {
            const data = rargs
                .concat(optional_1.OrUndefined(this.givenInitial))
                .filter((i) => optional_1.isSome(this.coerce(i)))
                .reduce((r, val) => {
                this.attributes.forEach((attr) => {
                    const m = attr.type.coerce(val[attr.name]);
                    if (optional_1.isSome(m)) {
                        r[attr.name].push(m.some);
                    }
                });
                return r;
            }, this.attributes.reduce((r, attr) => {
                r[attr.name] = [];
                return r;
            }, {}));
            return this.attributes.reduce((r, attr) => {
                r[attr.name] = attr.type.create(...data[attr.name]);
                return r;
            }, {});
        }
        toStreamChunk(val, chunk) {
            throw new Error('Method not implemented.');
        }
    }
    Definition.type = exports.typeName;
    return Definition;
})();
exports.Definition = Definition;
//# sourceMappingURL=struct.js.map