"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSImports = exports.TSImport = void 0;
class TSImport {
    constructor(imp) {
        this.imp = imp;
    }
    get key() {
        if (this.imp.external) {
            return this.imp.external.fname;
        }
        else {
            return this.imp.sWriter.fname;
        }
    }
    toString(wr) {
        if (this.imp.external) {
            const imports = this.imp.external.defs
                .sort((a, b) => {
                if (a.name < b.name) {
                    return 1;
                }
                else if (a.name > b.name) {
                    return -1;
                }
                return 0;
            })
                .map((def) => `${def.name}${def.as ? ` as ${def.as}` : ''}`)
                .join(', ');
            return `import { ${imports} } from ${wr.quote(this.imp.external.fname)};`;
        }
        else {
            return `import { ${this.imp.sWriter.def.name} } from ${wr.quote(this.imp.sWriter.fname)};`;
        }
    }
}
exports.TSImport = TSImport;
class TSImports {
    constructor(args) {
        this.args = args;
        this.imports = new Map();
        this.add(new TSImport({
            external: {
                defs: [
                    {
                        name: 'Types',
                    },
                    {
                        name: 'Align',
                    },
                    {
                        name: 'Optional',
                    },
                    {
                        name: 'Utils',
                    },
                    {
                        name: 'ChunkBuffer',
                    },
                    {
                        name: 'StreamBuffer',
                    },
                    {
                        name: 'NestedReadonly',
                    },
                    {
                        name: 'NestedPartial',
                    },
                ],
                fname: args.definitionPath || 'flatterbuf/definition',
            },
        }));
    }
    add(ts) {
        let f = this.imports.get(ts.key);
        if (!f) {
            this.imports.set(ts.key, ts);
        }
        else if (ts.imp.external) {
            throw Error(`TSImport double:${ts.key}`);
        }
        return ts;
    }
    values() {
        return Array.from(this.imports.values());
    }
    prepend(wl, def) {
        wl.prependLine(0, ``);
        this.values()
            .filter((i) => !(i.imp.sWriter && i.imp.sWriter.def == def))
            .reverse()
            .forEach((i) => wl.prependLine(0, i.toString(wl.wr)));
        wl.prependLine(0, `// generated ${def.name}`);
    }
}
exports.TSImports = TSImports;
//# sourceMappingURL=ts-imports.js.map