"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSStructWriter = void 0;
const definition_1 = require("../definition");
// import { Runtime } from '../runtime';
const ts_1 = require("./ts");
const ts_imports_1 = require("./ts-imports");
const optional_1 = require("../definition/optional");
const types_1 = require("../definition/types");
class TSStructWriter {
    constructor(def, level, args) {
        this.def = def;
        this.level = level;
        this.args = args;
        this.imports = new ts_imports_1.TSImports(args);
        this.fname = `${args.generationPath}${this.def.name.toLowerCase()}`;
    }
    writeStructAttribute(wr, attr) {
        const wl = new ts_1.TSWriteLine(wr);
        wl.writeLine(0, `{`);
        wl.writeLine(1, `name: ${wl.wr.quote(attr.name)},`);
        wl.writeLine(1, `ofs: ${attr.ofs},`);
        let initial = optional_1.NoneOption;
        if (optional_1.isSome(this.def.givenInitial)) {
            initial = optional_1.SomeOption(this.def.givenInitial.some[attr.name]);
        }
        wl.writeLine(1, `type: ${this.getTypeDefinition(wl.wr, attr.name, attr.type, initial)}`);
        wl.write(0, `}`);
        return wl.toString();
    }
    alignFuncs(wr, alignFuncs) {
        return `{ element: ${wr.quote(alignFuncs.element)}, overall: ${wr.quote(alignFuncs.overall)} }`;
    }
    arrayInitial(wr, length, vals) {
        return vals
            .slice(0, length)
            .map((i) => {
            switch (typeof i) {
                case 'number':
                    return `${i}`;
                case 'string':
                    return `${wr.quote(i)}`;
                case 'object':
                    return JSON.stringify(i);
                default:
                    return `[]`;
            }
        })
            .join(', ');
    }
    initialArg(wr, tdef, _initial) {
        const wl = new ts_1.TSWriteLine(wr);
        if (definition_1.Optional.isSome(tdef.givenInitial)) {
            wl.writeLine(0, '{');
            wl.writeLine(1, `initial: ${JSON.stringify(tdef.givenInitial.some)}`);
            wl.write(0, '}');
        }
        return wl.toString();
    }
    writeArrayDefinition(wr, tdef) {
        const adef = tdef;
        let valueType = '';
        if (types_1.isFixedArray(tdef)) {
            if (types_1.isFixedArray(adef.element)) {
                valueType = adef.element.type;
            }
            else {
                valueType = adef.element.type;
            }
            const tsArray = `Types.${adef.type}.Definition`;
            return tsArray;
        }
        return `Types.${tdef.type}.Definition`;
    }
    getTypeDefinition(wr, attrName, tdef, initial) {
        const wl = new ts_1.TSWriteLine(wr);
        switch (tdef.type) {
            case definition_1.Types.Boolean.Definition.type:
                wl.writeLine(0, `new Types.Boolean.Definition(${this.initialArg(wr, tdef, initial)})`);
                break;
            case definition_1.Types.Char.Definition.type:
            case definition_1.Types.Uint8.Definition.type:
            case definition_1.Types.Uint16.Definition.type:
            case definition_1.Types.Short.Definition.type:
            case definition_1.Types.Uint32.Definition.type:
            case definition_1.Types.Int.Definition.type:
            case definition_1.Types.Float.Definition.type:
            case definition_1.Types.Double.Definition.type:
                wl.writeLine(0, `new Types.${tdef.type}.Definition(${this.initialArg(wr, tdef, initial)})`);
                break;
            case definition_1.Types.Uint64.Definition.type:
            case definition_1.Types.Long.Definition.type:
                wl.writeLine(0, `new Types.${tdef.type}.Definition(${this.initialArg(wr, tdef, initial)})`);
                break;
            case definition_1.Types.BitStruct.Definition.type:
                const bdef = tdef;
                this.addTypeReference(wr, bdef);
                wl.writeLine(0, `new ${bdef.name}.Definition(${this.initialArg(wr, tdef, initial)})`);
                break;
            case definition_1.Types.Struct.Definition.type:
                const sdef = tdef;
                this.addTypeReference(wr, sdef);
                wl.write(1, `new ${sdef.name}.Definition()`);
                break;
            case definition_1.Types.FixedArray.Definition.type:
            case definition_1.Types.FixedCString.Definition.type:
                const adef = tdef;
                const typeName = this.writeArrayDefinition(wr, tdef);
                wl.writeLine(0, `new ${typeName}({`);
                wl.writeLine(1, `length: ${adef.length},`);
                wl.writeLine(1, `alignFuncs: ${this.alignFuncs(wr, adef.alignFuncs)},`);
                if (tdef.type !== definition_1.Types.FixedCString.Definition.type) {
                    wl.writeLine(1, `element: ${this.getTypeDefinition(wr, `Item.${attrName || tdef.type}`, adef.element, optional_1.NoneOption)},`);
                }
                if (definition_1.Optional.isSome(adef.givenInitial)) {
                    wl.writeLine(1, `initial: [${this.arrayInitial(wr, adef.length, adef.givenInitial.some)}]`);
                }
                wl.write(0, `})`);
                break;
            default:
                throw Error(`getTypeDefinition failed for: ${tdef.type}`);
        }
        return wl.toString();
    }
    addTypeReference(wr, def) {
        if (definition_1.Types.isScalar(def)) {
            if (def.type === definition_1.Types.BitStruct.Definition.type) {
                const bdef = def;
                const tsw = wr.bitStructClass(bdef, this.level + 1);
                this.imports.add(new ts_imports_1.TSImport({ sWriter: tsw.writer }));
            }
            return;
        }
        if (definition_1.Types.isFixedArray(def)) {
            const adef = def;
            this.addTypeReference(wr, adef.element);
            return;
        }
        if (definition_1.Types.isStruct(def)) {
            const sdef = def;
            const tsw = wr.structClass(sdef, this.level + 1);
            this.imports.add(new ts_imports_1.TSImport({ sWriter: tsw.writer }));
            return;
        }
        throw Error(`addTypeReferenced for unknown type:${def.type}`);
    }
    writeInterface(wl) {
        wl.writeLine(1, `export interface MutableType {`);
        const sdef = this.def;
        sdef.attributes.forEach((i) => {
            this.addTypeReference(wl.wr, i.type);
            wl.writeLine(2, `${ts_1.attributeDefinition(i)}: ${ts_1.typeDefinition(i.type, 'MutableType')};`);
        });
        wl.writeLine(1, '}');
        wl.writeLine(1, `export interface Type {`);
        sdef.attributes.forEach((i) => {
            this.addTypeReference(wl.wr, i.type);
            wl.writeLine(2, `readonly ${ts_1.attributeDefinition(i)}: ${ts_1.typeDefinition(i.type, 'Type')};`);
        });
        wl.writeLine(1, '}');
        wl.writeLine(1, 'export type ValueType = Type;');
        // wl.writeLine(1, 'export type FromStreamFN  = (rb: StreamBuffer, name?: string) => Type;');
    }
    writeTSObjectItem(wr, v) {
        if (Array.isArray(v)) {
            return [`[\n`, ...v.map((i) => this.writeTSObjectItem(wr, i)), `\n],`].join('');
        }
        else if (typeof v === 'object') {
            return this.writeTSObject(wr, v);
        }
        else if (typeof v === 'string') {
            return wr.quote(v) + ',';
        }
        return v.toString() + ',';
    }
    writeTSObject(wr, hl) {
        const mid = Array.from(Object.entries(hl))
            .filter((v) => v[1] !== undefined)
            .map((v) => this.writeTSObjectItem(wr, v[1]));
        return [`{\n`, ...mid, `\n},`].join('');
    }
    writeAttributes(wr) {
        const wl = new ts_1.TSWriteLine(wr);
        wl.writeLine(0, '// tslint:disable-next-line: typedef');
        wl.writeLine(0, `public static readonly AttributeByName = {`);
        this.def.attributes.forEach((attr) => {
            wl.writeLine(1, `${attr.name}: ${this.writeStructAttribute(wl.wr, attr)},`);
        });
        wl.writeLine(0, `};\n`);
        wl.writeLine(0, `public static readonly Attributes: Types.Struct.AttributeOfs<unknown>[] = [`);
        this.def.attributes.forEach((attr) => {
            wl.writeLine(1, `Definition.AttributeByName.${attr.name},`);
        });
        wl.writeLine(0, '];');
        wl.writeLine(0, this.writeStaticGivenInitial(wl.wr));
        return wl.toString();
    }
    writeFilterFunction(wr) {
        const wl = new ts_1.TSWriteLine(wr);
        wl.writeLine(0, 'public coerce(val?: NestedPartial<Type>):');
        wl.writeLine(1, 'Optional.Option<NestedPartial<Type>> {');
        wl.writeLine(1, `if (typeof val !== 'object') {`);
        wl.writeLine(2, 'return Optional.NoneOption;');
        wl.writeLine(1, '}');
        wl.writeLine(1, 'let ret: NestedPartial<MutableType> = {};');
        wl.writeLine(1, 'let found = false;');
        this.def.attributes.forEach((i) => {
            wl.writeLine(1, `const my${i.name} = this.attributeByName.${i.name}.type.coerce(val.${i.name}) as Optional.Option<${ts_1.typeDefinition(i.type, 'Type')}>;`);
            wl.writeLine(1, `if (Optional.isSome(my${i.name})) {`);
            wl.writeLine(2, `found = true;`);
            wl.writeLine(2, `ret.${i.name} = my${i.name}.some;`);
            wl.writeLine(1, `}`);
        });
        wl.writeLine(1, 'return found ? Optional.SomeOption(ret) : Optional.NoneOption;');
        wl.writeLine(0, '}');
        return wl.toString();
    }
    writeStaticGivenInitial(wr) {
        const wl = new ts_1.TSWriteLine(wr);
        wl.write(0, `public static readonly givenInitial: Optional.Option<NestedPartial<Type>>`);
        if (optional_1.isSome(this.def.givenInitial)) {
            wl.writeLine(0, ` = Optional.SomeOption(${ts_1.tsStringify(this.def.givenInitial.some, this.def, wr)});`);
        }
        else {
            wl.writeLine(0, ' = Optional.NoneOption;');
        }
        return wl.toString();
    }
    writeDefinition(wl) {
        wl.writeLine(1, `\nexport class Definition extends Types.Struct.AbstractDefinition {`);
        wl.writeLine(2, this.writeAttributes(wl.wr));
        // wl.writeLine(2, this.writeInitial(wl.wr));
        wl.writeLine(2, `public readonly type: Types.Base.TypeName = Types.Struct.Definition.type;`);
        wl.writeLine(2, `public readonly name: string = '${this.def.name}';`);
        wl.writeLine(2, `public readonly bytes: number = ${this.def.bytes};`);
        wl.writeLine(2, `public readonly alignFuncs: Align.Funcs<string> = { element: ${wl.wr.quote(this.def.alignFuncs.element)}, overall: ${wl.wr.quote(this.def.alignFuncs.overall)} };`);
        wl.writeLine(2, `public readonly attributes: typeof Definition.Attributes = Definition.Attributes;`);
        wl.writeLine(2, `public readonly attributeByName: typeof Definition.AttributeByName = Definition.AttributeByName;`);
        // wl.writeLine(2, `public readonly initial: Partial<Type>;`);
        wl.writeLine(2, `public readonly givenInitial: Optional.Option<NestedPartial<Type>>;`);
        wl.writeLine(2, ``);
        // wl.writeLine(2, 'public readonly fromStream: FromStreamFN = Definition.fromStream;');
        wl.writeLine(2, this.writeFromStream(wl.wr));
        wl.writeLine(2, this.writeCreateFunction(wl.wr));
        wl.writeLine(2, this.writeToStream(wl.wr));
        wl.writeLine(2, this.writeFilterFunction(wl.wr));
        wl.writeLine(2, `constructor(props?: { initial?: Partial<Type> }) {`);
        wl.writeLine(3, `super();`);
        wl.writeLine(3, `const my = (props || {}).initial;`);
        wl.writeLine(3, `this.givenInitial = Utils.nestedAssign(undefined, {},`);
        wl.writeLine(4, `Optional.OrUndefined(this.coerce(my)),`);
        wl.writeLine(4, `Optional.OrUndefined(Definition.givenInitial));`);
        // wl.writeLine(3, `this.initial = this.create(my);`);
        wl.writeLine(2, `}`);
        wl.writeLine(1, `}\n`);
    }
    getArrayLengths(def, ret = '') {
        if (def.type === definition_1.Types.FixedArray.Definition.type) {
            if (ret.length) {
                ret += ', ';
            }
            const adef = def;
            ret += `${adef.length}`;
            return this.getArrayLengths(adef.element, ret);
        }
        return ret;
    }
    writeCreateAttribute(_level, _attr, aname, vname) {
        return `Definition.AttributeByName.${aname}.type.create(${vname}) as ${ts_1.typeDefinition(_attr, 'Type')}`;
    }
    emptyNestedArray(type) {
        switch (definition_1.Types.toAttributeType(type)) {
            case definition_1.Types.AttributeType.FixedArray:
                const adef = type;
                if (definition_1.Types.toAttributeType(adef.element) !== definition_1.Types.AttributeType.FixedArray) {
                    return '[]';
                }
                return `[ ${Array(adef.length)
                    .fill(0)
                    .map((_, i) => this.emptyNestedArray(adef.element))
                    .join(', ')} ]`;
            case definition_1.Types.AttributeType.Scalar:
            case definition_1.Types.AttributeType.Struct:
                return '[]';
        }
    }
    writeCreateFunction(wr) {
        const wl = new ts_1.TSWriteLine(wr);
        // wl.writeLine(0, 'export function create(...rargs: Partial<Type>[]): Type {');
        wl.writeLine(0, 'public create(...rargs: NestedPartial<Type>[]): Type {');
        wl.writeLine(1, `const data = rargs.concat(Optional.OrUndefined(this.givenInitial))`);
        wl.writeLine(2, `.filter(i => Optional.isSome(this.coerce(i))).reduce((r, i) => {`);
        this.def.attributes.forEach((i) => {
            wl.writeLine(2, `if (i.${i.name} !== undefined) {`);
            wl.writeLine(3, `r.${i.name}.push(i.${i.name});`);
            wl.writeLine(2, `}`);
        });
        wl.writeLine(2, 'return r;');
        wl.writeLine(1, `}, {`);
        this.def.attributes.forEach((i) => {
            wl.writeLine(2, `${i.name}: ${this.emptyNestedArray(i.type)} as ${ts_1.typeDefinition(i.type, 'Type', (s) => `NestedPartial<${s}>`)}[],`);
            this.addTypeReference(wl.wr, i.type);
        });
        wl.writeLine(1, '});');
        wl.writeLine(1, `return {`);
        this.def.attributes.forEach((attr) => {
            wl.writeLine(2, `${attr.name}: ${this.writeCreateAttribute(0, attr.type, attr.name, `...data.${attr.name}`)},`);
        });
        wl.writeLine(1, `};`);
        wl.writeLine(0, '}');
        return wl.toString();
    }
    fromStreamAction(wr, attrName, def, stype, level = 0, itemName) {
        switch (definition_1.Types.toAttributeType(def)) {
            case definition_1.Types.AttributeType.Scalar:
                if (def.type === definition_1.Types.FixedCString.Definition.type) {
                    // const cdef = (def as unknown) as Types.FixedCString.Definition;
                    return `Definition.AttributeByName.${attrName}.${stype}.fromStreamChunk(nrb, ${wr.backQuote(itemName || attrName)}) as number[]`;
                }
                else if (def.type === definition_1.Types.BitStruct.Definition.type) {
                    // const bdef = (def as unknown) as Types.BitStruct.Definition;
                    return `Definition.AttributeByName.${attrName}.${stype}.fromStreamChunk(nrb, ${wr.backQuote(itemName || attrName)})`;
                }
                return `Definition.AttributeByName.${attrName}.${stype}.fromStreamChunk(nrb, ${wr.backQuote(itemName || attrName)})`;
            // return `nrb.read${def.type}()`;
            case definition_1.Types.AttributeType.Struct:
                // const sdef = def as Types.Base.NamedType<B>;
                return `Definition.AttributeByName.${attrName}.${stype}.fromStreamChunk(nrb, ${wr.backQuote(itemName || attrName)})`;
            case definition_1.Types.AttributeType.FixedArray:
                return `Definition.AttributeByName.${attrName}.${stype}.fromStreamChunk(nrb, ${wr.backQuote(itemName || attrName)}) as ${ts_1.typeDefinition(def, 'Type')}`;
            // return this.writeStreamArrayAction(wr, attrName, def, 'type.element', level, itemName || attrName);
        }
        throw Error(`fromStreamAction: ${def.type}`);
    }
    writeFromStream(wr) {
        const wl = new ts_1.TSWriteLine(wr);
        wl.writeLine(0, 'public static fromStreamChunk(nrb: ChunkBuffer, name: string = this.name): Type {');
        wl.writeLine(1, 'return {');
        this.def.attributes.forEach((i) => {
            wl.writeLine(2, `${i.name}: ${this.fromStreamAction(wl.wr, i.name, i.type, 'type')},`);
        });
        wl.writeLine(1, '};');
        wl.writeLine(0, '}');
        wl.writeLine(0, '');
        wl.writeLine(0, 'public static fromStream(rb: StreamBuffer, name: string = this.name): Type {');
        wl.writeLine(1, `return rb.prepareRead(name, ${this.def.bytes}, (nrb) => this.fromStreamChunk(nrb, name));`);
        wl.writeLine(0, '}');
        wl.writeLine(0, '');
        wl.writeLine(0, '// tslint:disable-next-line: typedef');
        wl.writeLine(0, 'public readonly fromStreamChunk = Definition.fromStreamChunk;');
        wl.writeLine(0, '');
        return wl.toString();
    }
    toStreamAction(level, wr, aname, vname, def) {
        switch (definition_1.Types.toAttributeType(def)) {
            case definition_1.Types.AttributeType.Scalar:
                if (def.type === definition_1.Types.BitStruct.Definition.type) {
                    // const bdef = (def as unknown) as Definition.Types.BitStruct;
                    return `Definition.AttributeByName.${aname}.type.toStreamChunk(${vname}, nwb, name)`;
                }
                else if (def.type === definition_1.Types.FixedCString.Definition.type) {
                    // const cdef = (def as unknown) as Types.FixedCString.Definition;
                    return `Definition.AttributeByName.${aname}.type.toStreamChunk(${vname}, nwb, name)`;
                }
                return `Definition.AttributeByName.${aname}.type.toStreamChunk(${vname}, nwb, name)`;
            // return `nwb.write${def.type}(${vname})`;
            case definition_1.Types.AttributeType.Struct:
                // const sdef = def as Types.Type.DefinitionNameAttribute<B>;
                // return `${sdef.name}.toStream(${aname}, nwb.sbuf)`;
                return `Definition.AttributeByName.${aname}.type.toStreamChunk(${vname}, nwb, name)`;
                break;
            case definition_1.Types.AttributeType.FixedArray:
                // const adef = (def as unknown) as Types.FixedArray.ArrayTypeAttribute<B>;
                return `Definition.AttributeByName.${aname}.type.toStreamChunk(${vname}, nwb, name)`;
            /*
                ${
                adef.length
              }, (idx${level}) => ${this.toStreamAction(
                level + 1,
                wr,
                'XXXXX',
                `(${vname} || [])[idx${level}]`,
                adef.element,
              )})`;
              */
        }
    }
    writeToStream(wr) {
        const wl = new ts_1.TSWriteLine(wr);
        wl.writeLine(0, 'public toStreamChunk(data: Partial<Type>, nwb: ChunkBuffer, name: string = this.name): void {');
        wl.writeLine(1, 'const tmp = this.create(data);');
        this.def.attributes.forEach((i) => {
            wl.writeLine(1, this.toStreamAction(0, wl.wr, i.name, `tmp.${i.name}`, i.type) + ';');
        });
        wl.writeLine(0, '}');
        wl.writeLine(0, '');
        wl.writeLine(0, 'public toStream(data: Partial<Type>, wb: StreamBuffer, name: string = this.name): StreamBuffer {');
        wl.writeLine(1, `return wb.prepareWrite(name, ${this.def.bytes}, (nwb) => this.toStreamChunk(data, nwb, name));`);
        wl.writeLine(0, '}');
        return wl.toString();
    }
    write(wr) {
        const wl = new ts_1.TSWriteLine(wr);
        wl.writeLine(0, `export namespace ${this.def.name} {`);
        this.writeInterface(wl);
        this.writeDefinition(wl);
        // writeCloneFunction(wl);
        // this.writeToStream(wl);
        wl.writeLine(0, '}');
        this.imports.prepend(wl, this.def);
        return wl;
    }
}
exports.TSStructWriter = TSStructWriter;
//# sourceMappingURL=ts-struct-writer.js.map