"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSGenerator = exports.TSWriter = exports.initialValue = exports.attributeDefinition = exports.typeDefinition = exports.tsStringify = exports.writeCloneFunction = exports.TSWriteLine = void 0;
// import { Definition } from '../definition';
const ts_struct_writer_1 = require("./ts-struct-writer");
const ts_bit_struct_writer_1 = require("./ts-bit-struct-writer");
const definition_1 = require("../definition");
const optional_1 = require("../definition/optional");
class TSWriteLine {
    constructor(wr) {
        this.wr = wr;
        this.lines = [];
    }
    prependLine(i, line) {
        this.action(i, line, true, Array.prototype.unshift);
    }
    prepend(i, line) {
        this.action(i, line, false, Array.prototype.unshift);
    }
    writeLine(i, lines) {
        this.action(i, lines, true, Array.prototype.push);
    }
    write(i, lines) {
        this.action(i, lines, false, Array.prototype.push);
    }
    action(i, lines, appendNewLine, action) {
        const indent = this.wr.indent(i);
        const args = lines.split('\n').map((line, idx, spnl) => {
            let trimmed = line.trimEnd();
            if (trimmed.length) {
                trimmed = indent + trimmed;
            }
            if (appendNewLine || idx < spnl.length - 1) {
                trimmed += '\n';
            }
            return trimmed;
        });
        // console.log(`action=>`, args)
        action.apply(this.lines, args);
    }
    toString() {
        const ret = this.lines.join('');
        this.toString = () => ret;
        return ret;
    }
}
exports.TSWriteLine = TSWriteLine;
function writeCloneFunction(wl) {
    wl.writeLine(1, `\nexport function clone(prop?: { initial?: Partial<Type> }): Type {`);
    wl.writeLine(2, `return Reflection.create(typeof prop === 'object' ? prop.initial : undefined);`);
    wl.writeLine(1, `}\n`);
}
exports.writeCloneFunction = writeCloneFunction;
function tsStringify(iobj, typ, wr) {
    const wl = new TSWriteLine(wr);
    const oval = typ.coerce(iobj);
    if (optional_1.isNone(oval)) {
        return wl.toString();
    }
    const val = oval.some;
    switch (typ.type) {
        case definition_1.Types.Boolean.Definition.type:
            wl.write(0, val ? 'true' : 'false');
            break;
        case definition_1.Types.Char.Definition.type:
        case definition_1.Types.Uint8.Definition.type:
        case definition_1.Types.Uint16.Definition.type:
        case definition_1.Types.Short.Definition.type:
        case definition_1.Types.Uint32.Definition.type:
        case definition_1.Types.Int.Definition.type:
        case definition_1.Types.Float.Definition.type:
        case definition_1.Types.Double.Definition.type:
            wl.write(0, `${val}`);
            break;
        case definition_1.Types.Uint64.Definition.type:
        case definition_1.Types.Long.Definition.type:
            const hl = val;
            const out = [];
            if (typeof hl.low === 'number') {
                out.push(`low: ${hl.low}`);
            }
            if (typeof hl.high === 'number') {
                out.push(`high: ${hl.high}`);
            }
            wl.write(0, `{ ${out.join(', ')} }`);
            break;
        case definition_1.Types.BitStruct.Definition.type:
            const bitStruct = typ;
            const bobj = iobj;
            wl.writeLine(0, `{`);
            bitStruct.bits.forEach((bits) => {
                const bval = bobj[bits.name];
                if (definition_1.Optional.isSome(bits.type.coerce(bval))) {
                    wl.writeLine(1, `${bits.name}: ${tsStringify(bval, bits.type, wr)},`);
                }
            });
            wl.write(0, `}`);
            break;
        case definition_1.Types.Struct.Definition.type:
            const struct = typ;
            const obj = iobj;
            wl.writeLine(0, `{`);
            struct.attributes.forEach((attr) => {
                const sval = obj[attr.name];
                if (definition_1.Optional.isSome(attr.type.coerce(sval))) {
                    wl.writeLine(1, `${attr.name}: ${tsStringify(sval, attr.type, wr)},`);
                }
            });
            wl.write(0, `}`);
            break;
        case definition_1.Types.FixedCString.Definition.type:
        // const stype = typeof val;
        // // wl.write(0, `/*${JSON.stringify(val)}*/`);
        // if (stype === 'string') {
        //   wl.write(0, wr.quote(val as string));
        // }
        // if (stype === 'number') {
        //   wl.write(0, `${val as number}`);
        // }
        // break;
        case definition_1.Types.FixedArray.Definition.type:
            const atype = typ;
            const o = val.map(i => tsStringify(i, atype.element, wr));
            wl.write(0, `[${o.join(', ')}]`);
            break;
        default:
            throw Error(`tsStringify failed for: ${typ.type}`);
    }
    return wl.toString();
}
exports.tsStringify = tsStringify;
function typeDefinition(t, typeName, wrapStruct = (s) => s) {
    switch (t.type) {
        case definition_1.Types.Boolean.Definition.type:
            return 'boolean';
        case definition_1.Types.Char.Definition.type:
        case definition_1.Types.Uint8.Definition.type:
        case definition_1.Types.Uint16.Definition.type:
        case definition_1.Types.Short.Definition.type:
        case definition_1.Types.Uint32.Definition.type:
        case definition_1.Types.Int.Definition.type:
        case definition_1.Types.Float.Definition.type:
        case definition_1.Types.Double.Definition.type:
            return 'number';
        case definition_1.Types.Uint64.Definition.type:
        case definition_1.Types.Long.Definition.type:
            return wrapStruct(`Types.${t.type}.${typeName}`);
        case definition_1.Types.BitStruct.Definition.type:
            return wrapStruct(`${t.name}.${typeName}`);
        case definition_1.Types.Struct.Definition.type:
            return wrapStruct(`${t.name}.${typeName}`);
        case definition_1.Types.FixedCString.Definition.type:
            return `number[]`;
        case definition_1.Types.FixedArray.Definition.type:
            return (typeDefinition(t.element, typeName, wrapStruct) + '[]');
        default:
            throw Error(`typeDefinition failed for: ${t.type}`);
    }
}
exports.typeDefinition = typeDefinition;
function attributeDefinition(def) {
    // return def.name + (def.notRequired ? '?' : '');
    return def.name;
}
exports.attributeDefinition = attributeDefinition;
function initialValue(_wr, vname, def) {
    switch (def.type) {
        case definition_1.Types.Uint64.Definition.type:
        case definition_1.Types.Long.Definition.type:
        // const hl = (def as Types.ScalarType<Runtime.Types.HighLow.Type>).initial;
        // return `{ high: ${hl.high}, low: ${hl.low} }`;
        case definition_1.Types.FixedCString.Definition.type:
        case definition_1.Types.BitStruct.Definition.type:
            return `dkdkfk:${def.type}:${vname}`;
        case definition_1.Types.Boolean.Definition.type:
        // return (def as Types.ScalarType<boolean>).initial.toString();
        case definition_1.Types.Char.Definition.type:
        case definition_1.Types.Uint8.Definition.type:
        case definition_1.Types.Uint16.Definition.type:
        case definition_1.Types.Short.Definition.type:
        case definition_1.Types.Uint32.Definition.type:
        case definition_1.Types.Int.Definition.type:
        case definition_1.Types.Float.Definition.type:
        case definition_1.Types.Double.Definition.type:
        // return (def as Types.ScalarType<number>).initial.toString();
        case definition_1.Types.Struct.Definition.type:
        case definition_1.Types.FixedArray.Definition.type:
            // return `Reflection.attributes[${wr.quote(sdef.name)}].initial`;
            return `${vname}`;
        default:
            throw Error(`initialValue failed for: ${vname} => ${def.type}`);
    }
}
exports.initialValue = initialValue;
class TSWriter {
    constructor(args = {}) {
        this.structs = new Map();
        this.fixedArrays = new Map();
        this.args = {
            indent: args.indent || '  ',
            ...args,
            quote: args.quote || "'",
            generationPath: args.generationPath || './',
            runtimePath: args.runtimePath || 'flatterbuf/runtime',
            definitionPath: args.definitionPath || 'flatterbuf/definition',
        };
    }
    indent(level) {
        return Array(level).fill(this.args.indent).join('');
    }
    quote(val) {
        return `${this.args.quote}${val.replace(new RegExp(this.args.quote, 'g'), `\\${this.args.quote}`)}${this.args.quote}`;
    }
    backQuote(val) {
        return `\`${val.replace(new RegExp('`', 'g'), `\\\``)}\``;
    }
    structClass(def, level) {
        let tsw = this.structs.get(def.name);
        if (!tsw) {
            tsw = { writer: new ts_struct_writer_1.TSStructWriter(def, level, this.args) };
            this.structs.set(tsw.writer.def.name, tsw);
        }
        return tsw;
    }
    bitStructClass(def, level) {
        let tsw = this.structs.get(def.name);
        if (!tsw) {
            tsw = { writer: new ts_bit_struct_writer_1.TSBitStructWriter(def, this.args) };
            this.structs.set(tsw.writer.def.name, tsw);
        }
        return tsw;
    }
    // public fixedArrayClass<B>(def: Types.FixedArray.Definition<B>, name: string) {
    //   let tsw = this.fixedArrays.get(def.name);
    //   if (!tsw) {
    //     tsw = { writer: new TSBitStructWriter(def, this.args) };
    //     this.structs.set(tsw.writer.def.name, tsw);
    //   }
    //   return tsw;
    // }
    generator(def) {
        if (definition_1.Types.isFixedArray(def)) {
            throw Error(`Implement-Array:${def.type}`);
        }
        else if (definition_1.Types.isStruct(def)) {
            if (def.type === definition_1.Types.Struct.Definition.type) {
                this.structClass(def, 0);
                return this;
            }
        }
        else if (definition_1.Types.isScalar(def)) {
            if (def.type === definition_1.Types.BitStruct.Definition.type) {
                this.bitStructClass(def, 0);
                return this;
            }
            throw Error(`Implement-Scalar:${def.type}`);
        }
        throw Error(`Unknown Type:${def.type}`);
    }
    // public getImports() {}
    getStructs() {
        let ret;
        let preSize = 0;
        do {
            // first pass
            preSize = this.structs.size;
            // console.log('getStruct-1=', preSize);
            ret = Array.from(this.structs.values()).map((i) => {
                if (!i.written) {
                    i.written = i.writer.write(this);
                }
                return i;
            });
        } while (this.structs.size != preSize);
        // console.log('getStruct-2=', preSize);
        return ret;
    }
}
exports.TSWriter = TSWriter;
function TSGenerator(def, writer = new TSWriter()) {
    return writer.generator(def);
}
exports.TSGenerator = TSGenerator;
//# sourceMappingURL=ts.js.map